# File Name: __init__.py
# Created By: ZW
# Created On: 2022-12-19
# Purpose: submodule package init for gobjects.core

# module imports
# ----------------------------------------------------------------------------
from .bed import Bed6, Bed12
from .bedpe import Bedpe
from .gtf import Gtf
from .intervals import Interval
from .intervalmap import IntervalMap